import csv
import json
import sys


if __name__ == '__main__':
    with open(sys.argv[1], 'r') as fp:
        all_fieldnames = [
            'bank_code', 'feature', 'name', 'postal_code', 'place', 'short_name', 'pan', 'bic',
            'check_digit_method', 'record_number', 'mod_number', 'tbd', 'successor_bank_code']
        data = csv.DictReader(fp, all_fieldnames, delimiter=';')

        fieldnames = ['bank_code', 'name', 'short_name', 'bic']
        cleaned = []
        for row in data:
            if not row['bank_code'].strip():
                continue
            for key, value in row.items():
                if key not in fieldnames:
                    del row[key]
                    continue
                row[key] = value.strip()
            row['country_code'] = 'DE'
            if row['bic']:
                cleaned.append(row)

    with open('schwifty/bank-registry.json', 'w') as fp:
        json.dump(cleaned, fp, indent=2)



from functools import partial
import re

import iso3166

from schwifty.common import Base
from schwifty import registry


_bic_re = re.compile(r'[A-Z]{4}[A-Z]{2}[A-Z0-9]{2}(?:[A-Z0-9]{3})?')


class BIC(Base):
    """The BIC object.

    Examples:

        You can either create a new BIC object by providing a code as text::

            >>> bic = BIC('GENODEM1GLS')
            >>> bic.country_code
            'DE'
            >>> bic.location_code
            'M1'
            >>> bic.bank_code
            'GENO'

        or by using the :meth:`from_bank_code` classmethod::

            >>> bic = BIC.from_bank_code('DE', '43060967')
            >>> bic.formatted
            'GENO DE M1 GLS'
    Args:
        bic (str): The BIC number.
        allow_invalid (bool): If set to ``True`` validation is skipped on instantiation.
    """

    def __init__(self, bic, allow_invalid=False):
        super(BIC, self).__init__(bic)
        if not allow_invalid:
            self.validate()

    @classmethod
    def from_bank_code(cls, country_code, bank_code):
        """Create a new BIC object from country- and bank-code.

        Args:
            country_code (str): ISO 3166 alpha2 country-code.
            bank_code (str): Country specific bank-code.

        Note:
            This currently only works for German bank-codes.
        """
        try:
            return cls(registry.get('bank_code')[(country_code, bank_code)]['bic'])
        except KeyError:
            pass

    def validate(self):
        self._validate_length()
        self._validate_structure()
        self._validate_country_code()
        return True

    def _validate_length(self):
        if self.length not in (8, 11):
            raise ValueError('Invalid length %d' % self.length)

    def _validate_structure(self):
        if not _bic_re.match(self.compact):
            raise ValueError('Invalid structure %s' % self.compact)

    def _validate_country_code(self):
        country_code = self.country_code
        try:
            iso3166.countries_by_alpha2[country_code]
        except KeyError:
            raise ValueError('Invalid country code %s' % country_code)

    @property
    def formatted(self):
        """str: The BIC separated in the blocks bank-, country- and location-code."""
        formatted = ' '.join([self.bank_code, self.country_code, self.location_code])
        if self.branch_code:
            formatted += ' ' + self.branch_code
        return formatted

    @property
    def country_bank_code(self):
        """str or None: The country specific bank-code associated to the BIC."""
        entry = registry.get('bic').get(self.compact)
        if entry:
            return entry.get('bank_code')

    @property
    def exists(self):
        """bool: Indicates if the BIC is available in Schwifty's registry."""
        return bool(registry.get('bic').get(self.compact))

    bank_code = property(partial(Base._get_component, start=0, end=4),
                         doc='str: The bank-code part of the BIC.')
    branch_code = property(partial(Base._get_component, start=8, end=11),
                           doc='str or None: The branch-code part of the BIC (if available)')
    country_code = property(partial(Base._get_component, start=4, end=6),
                            doc='str: The ISO 3166 alpha2 country-code.')
    location_code = property(partial(Base._get_component, start=6, end=8),
                             doc='str: The location code of the BIC.')


registry.build_index('bank', 'bic', 'bic')
registry.build_index('bank', 'bank_code', ('country_code', 'bank_code'))



import pytest

from schwifty import BIC


def test_bic():
    bic = BIC('GENODEM1GLS')
    assert bic.formatted == 'GENO DE M1 GLS'
    assert bic.validate()


def test_bic_allow_invalid():
    bic = BIC('GENODXM1GLS', allow_invalid=True)
    assert bic
    assert bic.country_code == 'DX'
    with pytest.raises(ValueError):
        bic.validate()


def test_bic_no_branch_code():
    bic = BIC('GENODEM1')
    assert bic.branch_code is None
    assert bic.formatted == 'GENO DE M1'


def test_country_bank_code():
    assert BIC('ABNAJPJTXXX').country_bank_code is None
    assert BIC('GENODEM1GLS').country_bank_code == '43060967'


def test_bic_properties():
    bic = BIC('GENODEM1GLS')
    assert bic.length == 11
    assert bic.bank_code == 'GENO'
    assert bic.branch_code == 'GLS'
    assert bic.country_code == 'DE'
    assert bic.location_code == 'M1'
    assert bic.country_bank_code == '43060967'
    assert bic.exists


@pytest.mark.parametrize('code', [
    'AAAA',             # Too short
    'AAAADEM1GLSX',     # Too long
    '12ABDEM1GLS',      # Wrong structure in banc-id
    'GENOD1M1GLS',      # Wrong structure in country-code
    'GENOXXM1GLS',      # Wrong country-code
])
def test_invalid_bic(code):
    with pytest.raises(ValueError):
        BIC(code)


def test_bic_from_bank_code():
    bic = BIC.from_bank_code('DE', '43060967')
    assert bic.compact == 'GENODEM1GLS'


def test_bic_from_unknown_bank_code():
    assert not BIC.from_bank_code('PO', '12345678')


def test_magic_methods():
    bic = BIC('GENODEM1GLS')
    assert bic == 'GENODEM1GLS'
    assert bic == BIC('GENODEM1GLS')
    assert bic != BIC('GENODEMMXXX')
    assert bic != 12345
    assert bic < 'GENODEM1GLT'

    assert str(bic) == 'GENODEM1GLS'
    assert repr(bic) == '<BIC=GENODEM1GLS>'



from functools import total_ordering
import re


_clean_regex = re.compile(r'\s+')


@total_ordering
class Base(object):

    def __init__(self, code):
        self._code = clean(code)

    def __str__(self):
        return self.compact

    def __repr__(self):
        return '<%s=%s>' % (self.__class__.__name__, str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    @property
    def compact(self):
        '''str: Compact representation of the code.'''
        return self._code

    @property
    def length(self):
        '''int: Length of the compact code.'''
        return len(self.compact)

    def _get_component(self, start, end=None):
        if start < self.length and (end is None or end <= self.length):
            return self.compact[start:end] if end else self.compact[start:]


def clean(string):
    return _clean_regex.sub('', string).upper()



import json
from pkg_resources import resource_filename


_registry = {}


def has(name):
    return name in _registry


def get(name):
    if not has(name):
        with open(resource_filename(__name__, name + '-registry.json'), 'r') as fp:
            save(name, json.load(fp))
    return _registry[name]


def save(name, data):
    _registry[name] = data


def build_index(base_name, index_name, key):
    def make_key(entry):
        return tuple(entry[k] for k in key) if isinstance(key, tuple) else entry[key]
    save(index_name, dict((make_key(entry), entry) for entry in get(base_name)))


def manipulate(name, func):
    registry = get(name)
    if isinstance(registry, dict):
        for key, value in registry.items():
            registry[key] = func(key, value)
    elif isinstance(registry, list):
        registry = [func(item) for item in registry]
    save(name, registry)



import csv
import json
import re

from bs4 import BeautifulSoup
import requests


url = 'https://www.swift.com/standards/data-standards/iban'


def get_raw():
    soup = BeautifulSoup(requests.get(url).content, 'html.parser')
    link = soup.find('a', attrs={'data-title': 'IBAN Registry TXT'})
    return requests.get(link['href']).content


def clean(raw):
    chars = ' \t\n\r;:\'"'
    return [{key.strip(chars).lower(): value.strip(chars) for key, value in line.items()}
            for line in raw]


def parse_positions(bban_length, spec):
    pattern = r'(.*?)(\d+)\s*(?:-|to)\s*(\d+)'
    matches = re.findall(pattern, spec)
    positions = {'bank_code': [0, 0], 'branch_code': [0, 0], 'account_code': [0, 0]}

    def match_to_range(match):
        return [int(match[1]) - 1, int(match[2])]

    positions['account_code'] = [0, bban_length]
    if matches:
        positions['bank_code'] = match_to_range(matches[0])
        positions['account_code'][0] = positions['bank_code'][1]
    if len(matches) > 1 and 'Branch' in matches[1][0]:
        positions['branch_code'] = match_to_range(matches[1])
        positions['account_code'][0] = positions['branch_code'][1]
    else:
        positions['branch_code'] = [positions['bank_code'][1], positions['bank_code'][1]]
    return positions


if __name__ == '__main__':
    raw = get_raw()
    data = csv.DictReader(raw.splitlines(), delimiter='\t', quotechar='"')
    registry = {}
    for row in clean(data):
        codes = re.findall(r'[A-Z]{2}', row['country code as defined in iso 3166'])
        entry = {
            'bban_spec': row['bban structure'],
            'bban_length': row['bban length'],
            'iban_spec': re.match(r'[A-Za-z0-9!]+', row['iban structure']).group(0),
            'iban_length': int(row['iban length']),
        }
        if entry['bban_spec'] == 'Not in use':
            entry['bban_spec'] = re.sub(r'[A-Z]{2}2!n', '', entry['iban_spec'])
        entry['bban_spec'] = entry['bban_spec'].replace(' ', '')

        if entry['bban_length'] == 'Not in use':
            entry['bban_length'] = entry['iban_length'] - 4
        else:
            entry['bban_length'] = int(entry['bban_length'])

        entry['positions'] = parse_positions(
            entry['bban_length'],
            row['bank identifier position within the bban'])

        for code in codes:
            registry[code] = entry

    with open('schwifty/iban-registry.json', 'w+') as fp:
        json.dump(registry, fp, indent=2)



# encoding: utf-8

from functools import partial
import re
import string

from schwifty.bic import BIC
from schwifty.common import Base
from schwifty import registry


_spec_to_re = {
    'n': r'\d',
    'a': r'[A-Z]',
    'c': r'[A-Za-z0-9]',
    'e': r' '
}

_alphabet = string.digits + string.ascii_uppercase


def _get_iban_spec(country_code):
    try:
        return registry.get('iban')[country_code]
    except KeyError:
        raise ValueError('Unknown country-code %s' % country_code)


def numerify(string):
    return int(''.join(str(_alphabet.index(c)) for c in string))


def code_length(spec, code_type):
    start, end = spec['positions'][code_type]
    return end - start


class IBAN(Base):
    """The IBAN object.

    Examples:

        You create a new IBAN object by supplying an IBAN code in text form. The IBAN
        is validated behind the scenes and you can then access all relevant components
        as properties::

            >>> iban = IBAN('DE89 3704 0044 0532 0130 00')
            >>> iban.account_number
            '0532013000'
            >>> iban.bank_code
            '37040044'
            >>> iban.country_code
            'DE'
            >>> iban.check_digits
            '89'


    Args:
        iban (str): The IBAN code.
        allow_invalid (bool): If set to `True` IBAN validation is skipped on instantiation.
    """

    def __init__(self, iban, allow_invalid=False):
        super(IBAN, self).__init__(iban)
        if self.checksum_digits == '??':
            self._code = self.country_code + self._calc_checksum_digits() + self.bban

        if not allow_invalid:
            self.validate()

    def _calc_checksum_digits(self):
        return '{:2d}'.format(98 - (numerify(self.bban + self.country_code) * 100) % 97)

    @classmethod
    def generate(cls, country_code, bank_code, account_code):
        """Generate an IBAN from it's components.

        If the bank-code and/or account-number have less digits than required by their
        country specific representation, the respective component is padded with zeros.

        Examples:

            To generate an IBAN do the following::

                >>> bank_code = '37040044'
                >>> account_code = '532013000'
                >>> iban = IBAN.generate('DE', bank_code, account_code)
                >>> iban.formatted
                'DE89 3704 0044 0532 0130 00'

        Args:
            country_code (str): The ISO 3166 alpha-2 country code.
            bank_code (str): The country specific bank-code.
            account_code (str): The customer specific account-code.
        """
        spec = _get_iban_spec(country_code)
        bank_code_length = code_length(spec, 'bank_code')
        branch_code_length = code_length(spec, 'branch_code')
        bank_and_branch_code_length = bank_code_length + branch_code_length
        account_code_length = code_length(spec, 'account_code')

        if len(bank_code) > bank_and_branch_code_length:
            raise ValueError('Bank code exceeds maximum size %d' % bank_and_branch_code_length)

        if len(account_code) > account_code_length:
            raise ValueError('Account code exceeds maximum size %d' % account_code_length)

        bank_code = bank_code.rjust(bank_and_branch_code_length, '0')
        account_code = account_code.rjust(account_code_length, '0')
        iban = country_code + '??' + bank_code + account_code
        return cls(iban)

    def validate(self):
        self._validate_characters()
        self._validate_length()
        self._validate_format()
        self._validate_checksum()
        return True

    def _validate_characters(self):
        if not re.match(r'[A-Z]{2}\d{2}[A-Z]*', self.compact):
            raise ValueError('Invalid characters in IBAN')

    def _validate_checksum(self):
        if self.numeric % 97 != 1:
            raise ValueError('Invalid checksum digits')

    def _validate_length(self):
        if self.spec['iban_length'] != self.length:
            raise ValueError('Invalid IBAN length')

    def _validate_format(self):
        if not self.spec['regex'].match(self.bban):
            raise ValueError('Invalid BBAN structure: \'%s\' doesn\'t match %s' %
                             (self.bban, self.spec['bban_spec']))

    @property
    def numeric(self):
        """int: A numeric represenation of the IBAN."""
        return numerify(self.bban + self.compact[:4])

    @property
    def formatted(self):
        """str: The IBAN formatted in blocks of 4 digits."""
        return ' '.join(self.compact[i:i + 4] for i in range(0, len(self.compact), 4))

    @property
    def spec(self):
        """dict: The country specific IBAN specification."""
        return _get_iban_spec(self.country_code)

    @property
    def bic(self):
        """BIC: The BIC associated to the IBAN´s bank-code."""
        return BIC.from_bank_code(self.country_code, self.bank_code)

    def _get_code(self, code_type):
        start, end = self.spec['positions'][code_type]
        return self.bban[start:end]

    bban = property(partial(Base._get_component, start=4),
                    doc='str: The BBAN part of the IBAN.')
    country_code = property(partial(Base._get_component, start=0, end=2),
                            doc='str: ISO 3166 alpha-2 country code.')
    checksum_digits = property(partial(Base._get_component, start=2, end=4),
                               doc='str: Two digit checksum of the IBAN.')
    bank_code = property(partial(_get_code, code_type='bank_code'),
                         doc='str: The country specific bank-code.')
    branch_code = property(partial(_get_code, code_type='branch_code'),
                           doc='str or None: The branch-code of the bank if available.')
    account_code = property(partial(_get_code, code_type='account_code'),
                            doc='str: The customer specific account-code')


def add_bban_regex(country, spec):
    bban_spec = spec['bban_spec']
    spec_re = r'(\d+)(!)?([{}])'.format(''.join(_spec_to_re.keys()))

    def convert(match):
        quantifier = ('{%s}' if match.group(2) else '{1,%s}') % match.group(1)
        return _spec_to_re[match.group(3)] + quantifier
    spec['regex'] = re.compile('^%s$' % re.sub(spec_re, convert, bban_spec))
    return spec

registry.manipulate('iban', add_bban_regex)
